
from tnorm.version import version, __version__

from tnorm.TN_wrapper import TN_wrapper as load

import tnorm.kernel
import tnorm.extras
from tnorm.test import test
#import tnorm.GUI


#IN_SAGE=False
